/***************************************************************************
         evaluation.cpp  -  evaluates positions with network.dat file
                             -------------------
    begin                : Sat Jun 3 2000
    copyright            : (C) 1996-2001 by Matthias Luescher
    email                : m.luescher@datacomm.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <cassert>
#include <memory>
#include <math.h>

#include "def_inc.h"
#ifdef EN_VERSION
#include "myvector.h"
#endif // EN_VERSION
#include "engineobject.h"
#include "listclasses.h"
#include "pieces.h"

#include "evaluation.h"

#include <string>
#include <fstream>
#include <iostream>

using namespace std;

ConfigurationSet Evaluation::configs[MAX_POSITIONTYPES][6];
bool Evaluation::staticInitDone = false;

Evaluation::Evaluation()
{ 
	if ( !staticInitDone )
	{
		staticInitDone = true;
		LoadData(); 
	}
	currentConfigs = configs[OPENING]; 
	evalWindowOffset[0] = 0;
	evalWindowOffset[1] = 0;
#ifdef EN_VERSION
	lastGradSquare = 1;
#endif // EN_VERSION
};

int Evaluation::SetPositionType(int posType, bool clear)
{
int tpPieceCount[] = {0,0,0,0,0,0};
int iePieceCount[] = {0,0,0,0,0,0};
int qualitySum = 0;

	if ( posType == AUTO_CHOOSE_POSITION_TYPE )
	{
	Piece* p = m_pKing[m_move.toPlay];
		do
		{
			tpPieceCount[p->type]++;
			qualitySum += p->value;
			p = p->next;
		} while ( p->type != KING );

		p = m_pKing[!m_move.toPlay];
		do
		{
			iePieceCount[p->type]++;
			qualitySum += p->value;
			p = p->next;
		} while ( p->type != KING );

		m_pawnEndgame = false;

		if ( qualitySum == 2 * Q_KING + Q_ROOK &&
			 (tpPieceCount[ROOK] == 1 || iePieceCount[ROOK] == 1 ) )
			m_positionType = KRK;
		else if ( qualitySum == 2 * Q_KING + Q_PAWN &&
			 (tpPieceCount[PAWN] == 1 || iePieceCount[PAWN] == 1 ) )
		{
			m_positionType = KPK;
			m_pawnEndgame = true;
		}
		else if ( qualitySum == 2 * Q_KING + Q_QUEEN &&
				 (tpPieceCount[QUEEN] == 1 || iePieceCount[QUEEN] == 1 ) )
			m_positionType = KQK;
		else if ( qualitySum == 2 * Q_KING + 2 * Q_BISHOP &&
				 (tpPieceCount[BISHOP] == 2 || iePieceCount[BISHOP] == 2 ) )
			m_positionType = KBBK;
		else if ( qualitySum == 2 * Q_KING + (tpPieceCount[PAWN] + iePieceCount[PAWN]) * Q_PAWN  
				&& tpPieceCount[PAWN] + iePieceCount[PAWN] > 1 )
		{
			m_positionType = KPXKPX;
			m_pawnEndgame = true;
		}
		else if ( qualitySum == 2 * Q_KING + Q_BISHOP + Q_KNIGHT &&
				  ((tpPieceCount[BISHOP] == 1 && tpPieceCount[KNIGHT] == 1) ||
				   (iePieceCount[BISHOP] == 1 && iePieceCount[KNIGHT] == 1)) )
		{
			m_positionType = KBNK;
		}
		else if ( qualitySum < Q_ENDGAME_1 ) m_positionType = ENDGAME_1;
		else if ( qualitySum < Q_ENDGAME_2 ) m_positionType = ENDGAME_2;
		else if ( qualitySum < Q_MIDGAME_1 ) m_positionType = MIDGAME_1;
		else if ( qualitySum < Q_MIDGAME_2 ) m_positionType = MIDGAME_2;
		else if ( qualitySum < Q_MIDGAME_3 ) m_positionType = MIDGAME_3;
		else if ( qualitySum < Q_MIDGAME_4 ) m_positionType = MIDGAME_4;
		else if ( qualitySum < Q_MIDGAME_5 ) m_positionType = MIDGAME_5;
		else m_positionType = OPENING;
	}
	else 
	{
		m_pawnEndgame = false;
		m_positionType = posType;
		switch ( posType )
		{
		case KPK: m_pawnEndgame = true; break;
		case KPXKPX: m_pawnEndgame = true; break;
		}
	}

	if ( clear )
	{
		tpBitcodeTable.Clear();
		ieBitcodeTable.Clear();
	}

	currentConfigs = configs[m_positionType];

	return m_positionType;
};

void Evaluation::SetEvaluationWindowOffset()
{
int value = GetPositionValue(evalWindowOffset[m_move.toPlay], 
							 evalWindowOffset[m_move.toPlay], 0);
	evalWindowOffset[m_move.toPlay] = value;
	evalWindowOffset[!m_move.toPlay] = -value;
};

// Control numbers
#define CN_PIECE 0
#define CN_POS_TYPE 1
#define CN_EOF 2

void Evaluation::LoadData()
{
// LoadData() and SaveData() will only work properly on a 32bit machine
fstream infile;
Configuration c;
int posType, piece;
float weight;
unsigned int in;

	infile.open("network.dat", ios::in | ios::binary);

	if ( infile.good() )
	{
		infile.read((char*) &in, sizeof(unsigned int));
		if ( in != FILE_VERSION )
		{
#ifdef EN_VERSION
			cout << "Error: wrong version of the network.dat file!\n";
#else // !EN_VERSION
			cout << "telluser Chessterfield has not the right version of the network.dat file!" << endl;
#endif
			infile.close();
			return;
		}
		
		posType = 0;
		piece = 0;

		infile.read((char*) &in, sizeof(unsigned int));
		while ( infile.good() )
		{
			if ( in == 0 )
			{
				infile.read((char*) &in, sizeof(unsigned int));
				switch ( in )
				{
				case CN_PIECE: infile.read((char*) &piece, sizeof(int)); break;
				case CN_POS_TYPE: infile.read((char*) &posType, sizeof(int)); break;
				case CN_EOF: infile.close(); return; break;
				}
			}
			else
			{
				c.bitcode = in;
				infile.read((char*) &c.not_msac, sizeof(unsigned int));
				infile.read((char*) &weight, sizeof(float));
#ifdef EN_VERSION
				c.weight = (WEIGHTTYPE) weight;
#else
				if ( weight > 0 )
					c.weight = (int) (10.0 * weight + 0.5); 
				else
					c.weight = (int) (-((-10.0 * weight) + 0.5));
#endif
				infile.read((char*) &c.jump, sizeof(int));
				configs[posType][piece].push_back(c);
			}
			infile.read((char*) &in, sizeof(unsigned int));
		}
	}
	else
	{
#ifdef EN_VERSION
		cout << "Error: network.dat file not found!\n";
#else // !EN_VERSION
		cout << "telluser Chessterfield could not find its brain (file network.dat)!" << endl;
#endif 
	}
	
	infile.close();
}
	
/*void Evaluation::LoadData()
{
fstream infile;
Configuration c;
int posType, piece;
string str;

	infile.open("network.dat", ios::in);

	if ( infile.good() )
	{
		posType = 0;
		piece = 0;

		infile >> str;
		while ( infile.good() )
		{
			if ( str == "PIECE" )
			{
				infile >> piece;
				infile >> str; 
			}
			else if ( str == "POSITION_TYPE" )
			{
				infile >> posType;
				configs[posType][piece].reserve(100);
			}
			else if ( str == "<" )
			{
				unsigned int bit;
				c.bitcode = 0;
				infile >> bit;
				while ( infile.good() )
				{
					c.bitcode |= BIT_0 << bit; 
					infile >> bit;
				}
				infile.clear();
				infile >> str;

				c.not_msac = 0;
				infile >> bit;
				while ( infile.good() )
				{
					c.not_msac |= BIT_0 << bit; 
					infile >> bit;
				}
				infile.clear();
				infile >> str;

				infile >> c.weight;
				infile >> c.jump;
				configs[posType][piece].push_back(c);
			}
			infile >> str;
		}
	}
	else
	{
#ifdef EN_VERSION
		cout << "Error: network.dat file not found!\n";
#else // !EN_VERSION
		cout << "telluser Chessterfield could not find its brain (file network.dat)!" << endl;
#endif
	}
	
	infile.close();
}*/

const bool kingFields[64] = {
	false, false, false, false, false, false, false, false, 
	false, false, false, false, false, false, false, false, 
	false, false, false, false, false, false, false, false,
	false, false, false, false, false, false, false, false, 
	false, false, false, false, false, false, false, false, 
	false, false, false, false, false, false, false, false, 
	false, false, false, false, false, false, false, false, 
	false, false, false, false, false, false, false, false};

void Evaluation::GenerateAtomics()
{
	// ------------------ Initialize atomic features ------------------------
Piece* p = m_pKing[m_move.toPlay];
unsigned int i1 = 0, i2 = 0;
int *tpPC, *iePC;
int branchValue = 0;
	memcpy(m_wPawnColumns, m_initPawnColumns, 10 * sizeof(int));
	memcpy(m_bPawnColumns, m_initPawnColumns, 10 * sizeof(int));

	if ( p->color == WHITE )
	{
		tpPC = m_wPawnColumns;
		iePC = m_bPawnColumns;
	}
	else
	{
		tpPC = m_bPawnColumns;
		iePC = m_wPawnColumns;
	}

	do
	{
		p->atomics = i1; //wird in evaluateAtomics() verwendet
		branchValue += p->value;
		if ( p->type == PAWN ) tpPC[(p->square & 7) + 1]++;
		p = p->next;
	} while ( p->type != KING );

	p = m_pKing[!m_move.toPlay];
	i2 = BIT_16;
	do
	{
		p->atomics = i2; //wird in evaluateAtomics() verwendet
		branchValue -= p->value;
		if ( p->type == PAWN ) iePC[(p->square & 7) + 1]++;
		p = p->next;
	} while ( p->type != KING );

// ------------------ Evaluate atomic features ---------------------------
	// Koenig wird wegen der Initialisierung der Koenigsfelder vorgezogen
	memcpy(m_kingFields1, kingFields, 64 * sizeof(bool));
	memcpy(m_kingFields2, kingFields, 64 * sizeof(bool));

	m_owKingFields = m_kingFields1;
	m_opKingFields = m_kingFields1;

	if ( branchValue < -500 ) i2 = BIT_0;
	else if ( branchValue > 500 ) i2 = BIT_1;
	else i2 = 0;
	p = m_pKing[!m_move.toPlay];
	p->atomics |= i2;
	p->evaluateAtomics();
	
	m_owKingFields = m_kingFields2;

	if ( branchValue > 500 ) i1 = BIT_0;
	else if ( branchValue < -500 ) i1 = BIT_1;
	else i1 = 0;
	p = m_pKing[m_move.toPlay];
	do 
	{
		p->atomics |= i1;
		p->evaluateAtomics();
		p = p->next;
	} while ( p->type != KING );

	m_owKingFields = m_kingFields1;
	m_opKingFields = m_kingFields2;

	p = m_pKing[!m_move.toPlay]->next;
	while ( p->type != KING )
	{
		p->atomics |= i2;
		p->evaluateAtomics();
		p = p->next;
	}

	// Distance between the kings
int k1 = m_pKing[0]->square;
int k2 = m_pKing[1]->square;
int dx = abs((k1 >> 3) - (k2 >> 3));
int dy = abs((k1 & 7) - (k2 & 7)); 

	switch ( dx )
	{
	case 0: m_pKing[0]->atomics |= BIT_8; m_pKing[1]->atomics |= BIT_8; break;
	case 1: m_pKing[0]->atomics |= BIT_15; m_pKing[1]->atomics |= BIT_15; break;
	case 2: m_pKing[0]->atomics |= BIT_21; m_pKing[1]->atomics |= BIT_21; break;
	case 3: m_pKing[0]->atomics |= BIT_22; m_pKing[1]->atomics |= BIT_22; break;
	}
	switch ( dy )
	{
	case 0: m_pKing[0]->atomics |= BIT_23; m_pKing[1]->atomics |= BIT_23; break;
	case 1: m_pKing[0]->atomics |= BIT_24; m_pKing[1]->atomics |= BIT_24; break;
	case 2: m_pKing[0]->atomics |= BIT_25; m_pKing[1]->atomics |= BIT_25; break;
	case 3: m_pKing[0]->atomics |= BIT_26; m_pKing[1]->atomics |= BIT_26; break;
	}
}

int Evaluation::GetPositionValue(int alpha, int beta, int branchValue)
{
// ------ "Exact" calculation only within a certain window ---------------
#ifndef EN_VERSION
	if ( branchValue + Q_EVALUATION_WINDOW + evalWindowOffset[m_move.toPlay] < alpha || 
		 branchValue - Q_EVALUATION_WINDOW + evalWindowOffset[m_move.toPlay] > beta )
		return branchValue;
#endif


WEIGHTTYPE value = 0, temp;
ConfigurationSet* d;
ConfigurationSetIt end, it;

	GenerateAtomics();

// ------------------ Calculate position value ---------------------------

Piece* p = m_pKing[m_move.toPlay];
	do
	{
		if ( !tpBitcodeTable.GetEntry(p->atomics, p->type, value) )
		{
			temp = 0;
			d = &currentConfigs[p->type];
			it = d->begin(); end = d->end();
			while ( it < end )
			{
				if ( it->IsActive(p->atomics) ) 
				{
					temp += it->weight;
					it += it->Jump(p->atomics);
				}
				else it += it->jump;
			}
			value += temp;
			tpBitcodeTable.AddEntry(p->atomics, p->type, temp);
		}

		p = p->next;
	} while ( p->type != KING );

	p = m_pKing[!m_move.toPlay];
	do
	{
		if ( !ieBitcodeTable.GetEntry(p->atomics, p->type, value) )
		{
			temp = 0;
			d = &currentConfigs[p->type];
			it = d->begin(); end = d->end();
			while ( it < end )
			{
				if ( it->IsActive(p->atomics) ) 
				{
					temp -= it->weight;
					it += it->Jump(p->atomics);
				}
				else it += it->jump;
			}
			value += temp;
			ieBitcodeTable.AddEntry(p->atomics, p->type, temp);
		}

		p = p->next;
	} while ( p->type != KING );

	if ( m_pKing[m_move.toPlay]->m_bCastled ) value += 5; // prefer castling
	if ( m_pKing[!m_move.toPlay]->m_bCastled ) value -= 5;
	
#ifdef EN_VERSION
	return branchValue + ((int) (10 * value));
#else
	return branchValue + value;
#endif
}

#ifdef EN_VERSION

/*void Evaluation::SaveData() // old version: save configurations as text file
{
fstream outfile;
int posType, bit;
ConfigurationSet* d;
ConfigurationSetIt end, it;

	outfile.open("network.dat", ios::out);

	if ( outfile.good() )
	{
		for ( int piece = 0; piece < 6; piece++ )
		{
			switch ( piece )
			{
			case 0: outfile << "\n\nPIECE 0 (PAWN)\n"; break;
			case 1: outfile << "\n\nPIECE 1 (KNIGHT)\n"; break;
			case 2: outfile << "\n\nPIECE 2 (BISHOP)\n"; break;
			case 3: outfile << "\n\nPIECE 3 (ROOK)\n"; break;
			case 4: outfile << "\n\nPIECE 4 (QUEEN)\n"; break;
			case 5: outfile << "\n\nPIECE 5 (KING)\n"; break;
			}

			for ( posType = 0; posType < MAX_POSITIONTYPES; posType++ )
			{
				outfile << "POSITION_TYPE " << posType << "\n";

				d = &configs[posType][piece];
				end = d->end();

				for( it = d->begin(); it < end; it++ ) 
				{
					outfile << "< ";
					for ( bit = 0; bit < 32; bit++ )
						if ( it->bitcode & (BIT_0 << bit) ) outfile << bit << " ";
					outfile << ">< ";
					for ( bit = 0; bit < 32; bit++ )
						if ( it->not_msac & (BIT_0 << bit) ) outfile << bit << " ";
					outfile << "> " << it->weight << " " << it->jump << "\n";
				}
			}
		}
	}
	else
	{
		cout << "Error: network.dat file not found.\n";
	}
	
	outfile.close();
}*/

void Evaluation::SaveData()
{
fstream outfile;
int posType, piece;
float weight;
unsigned int out;
ConfigurationSet* d;
ConfigurationSetIt end, it;

	outfile.open("network.dat", ios::out | ios::trunc | ios::binary );

	if ( outfile.good() )
	{
		out = FILE_VERSION;
		outfile.write((char*) &out, sizeof(unsigned int));

		for ( piece = 0; piece < 6; piece++ )
		{
			out = 0;
			outfile.write((char*) &out, sizeof(unsigned int));
			out = CN_PIECE;
			outfile.write((char*) &out, sizeof(unsigned int));
			outfile.write((char*) &piece, sizeof(int));

			for ( posType = 0; posType < MAX_POSITIONTYPES; posType++ )
			{
				out = 0;
				outfile.write((char*) &out, sizeof(unsigned int));
				out = CN_POS_TYPE;
				outfile.write((char*) &out, sizeof(unsigned int));
				outfile.write((char*) &posType, sizeof(int));

				d = &configs[posType][piece];
				end = d->end();

				for( it = d->begin(); it < end; it++ ) 
				{
					weight = (float) it->weight;
					outfile.write((char*) &it->bitcode, sizeof(unsigned int));
					outfile.write((char*) &it->not_msac, sizeof(unsigned int));
					outfile.write((char*) &weight, sizeof(float));
					outfile.write((char*) &it->jump, sizeof(int));
				}
			}
		}

		out = CN_EOF;
		outfile.write((char*) &out, sizeof(unsigned int));
	}
	else
	{
		cout << "Error: network.dat file not found.\n";
	}
	
	outfile.close();
}

void Evaluation::InitWeights()
{
ConfigurationSet* d;
ConfigurationSetIt end, it;
	for ( int type = 0; type < 6; type++ )
	{
		d = &currentConfigs[type]; 
		end = d->end();
		for( it = d->begin(); it < end; it++ ) it->weight = 0;
	}
}

double Evaluation::CGstep(PositionSet& set, bool output)
{
ConfigurationSet* d;
ConfigurationSetIt cend, cit;
PositionSetIt pstart, pend, pit;
double gradSquare = 0, temp, error, alpha, beta; 
double counter = 0, denominator = 0, msError = 0, averageError = 0;
int type;

	// Calculate gradient
	pstart = set.begin(); pend = set.end();
	for( pit = pstart; pit != pend; pit++ )
	{
		error = pit->value - GetExampleValue(*pit);
		msError += error * error;
		averageError += fabs(error);
		if ( !output ) UpdateGradient( *pit, 2 * error ); 
	}

	if ( output )
	{
		if ( set.size() > 0 )
		{
			cout << "f(w)=" << msError / (double) set.size() << "  Mean error: " 
				 << averageError / (double) set.size() <<"\n\n";
		}
		tpBitcodeTable.Clear();
		ieBitcodeTable.Clear();
		return 0;
	}
	else 
	{
		cout << ".";
		cout.flush();
	}

	// Calculate beta
	for ( type = 0; type < 6; type++ )
	{
		d = &currentConfigs[type];
		cend = d->end();
		for( cit = d->begin(); cit < cend; cit++ )
			gradSquare += cit->ngrad1 * cit->ngrad1 * cit->preconditioner;
	}

	if ( lastGradSquare == 0 ) beta = 0;
	else beta = gradSquare / lastGradSquare;
	lastGradSquare = gradSquare;

	// Get the new direction
	for ( type = 0; type < 6; type++ )
	{
		d = &currentConfigs[type];
		cend = d->end();
		for( cit = d->begin(); cit < cend; cit++ ) 
			cit->ngrad1 = cit->ngrad1 + beta * cit->ngrad0;
	}

	// Get the step size alpha
	pstart = set.begin(); pend = set.end();
	for( pit = pstart; pit < pend; pit++ ) // += 10 )
	{
		error = pit->value - GetExampleValue(*pit, temp);
		counter += error * temp;
		denominator += temp * temp;
	}
	if ( denominator == 0 ) alpha = 0;
	else alpha = counter / denominator;

	// Update the new weight vector 
	for ( type = 0; type < 6; type++ )
	{
		d = &currentConfigs[type];
		cend = d->end();
		for( cit = d->begin(); cit < cend; cit++ ) 
		{
			cit->weight += alpha * cit->ngrad1 * cit->preconditioner;
			cit->ngrad0 = cit->ngrad1;
			cit->ngrad1 = 0;
		}
	}

	// Values in the table are no longer valid
	tpBitcodeTable.Clear();
	ieBitcodeTable.Clear();

	if ( set.size() == 0 )
		return 0;
	else
		return msError / (double) set.size();
}

double Evaluation::Gstep(PositionSet& set)
{
ConfigurationSet* d;
ConfigurationSetIt cend, cit;
PositionSetIt pstart, pend, pit;
double denominator = 0, counter = 0, temp, error, msError = 0, averageError = 0, alpha;
int type;

	cout << "Building preconditioner ...";
	// Reset gradient, count and preconditioner of each configuration
	for ( type = 0; type < 6; type++ )
	{
		d = &currentConfigs[type];
		cend = d->end();
		for( cit = d->begin(); cit < cend; cit++ ) 
		{
			cit->count = 0;
			cit->ngrad1 = 0;
			cit->preconditioner = 0;
		}
	}

	// Calculate gradient
	pstart = set.begin(); pend = set.end();
	for( pit = pstart; pit != pend; pit++ )
	{
		error = (pit->value - BuildPreconditioner(*pit));
		msError += error * error;
		averageError += fabs(error);
		UpdateGradient(*pit, 2 * error);
	}

	// Postprocess the preconditioner "matrix"
	for ( type = 0; type < 6; type++ )
	{
		d = &currentConfigs[type];
		cend = d->end();
		for( cit = d->begin(); cit < cend; cit++ ) 
		{
			if ( cit->preconditioner != 0 ) cit->preconditioner = 1 / cit->preconditioner;
			else cit->preconditioner = 0; // Should be impossible
		}
	}
	cout << " done.\n\n";

	if ( set.size() > 0 )
	{
	   cout << "f(w)=" << msError / (double) set.size() << "  Mean error: " 
	   	    << averageError / (double) set.size() <<"\n";
	}

	// Calculate the step size alpha
	pstart = set.begin(); pend = set.end();
	for( pit = pstart; pit < pend; pit++ ) // += 10 )
	{
		error = pit->value - GetExampleValue(*pit, temp);
		counter += error * temp;
		denominator += temp * temp;
	}

	if ( denominator == 0 ) alpha = 0;
	else alpha = counter / denominator;

	// Update the weight vector
	lastGradSquare = 0;
	for ( type = 0; type < 6; type++ )
	{
		d = &currentConfigs[type];
		cend = d->end();
		for( cit = d->begin(); cit < cend; cit++ ) 
		{
			cit->weight += alpha * cit->ngrad1 * cit->preconditioner;
			lastGradSquare += cit->ngrad1 * cit->ngrad1 * cit->preconditioner;
			cit->ngrad0 = cit->ngrad1; //xx
			cit->ngrad1 = 0; //xx
		}
	}

	// Values in bitcode table are no longer valid
	tpBitcodeTable.Clear();
	ieBitcodeTable.Clear();

	if ( set.size() == 0 )
		return 0;
	else
		return msError  / (double) set.size();
}


float Evaluation::GetExampleValue(Position& pos)
{
ConfigurationSet* d;
ConfigurationSetIt cend, cit;
PiecePropertiesSetIt eend, eit;
float value = 0, temp;

	// Calculate position value
	eend = pos.tpPieces.end();
	for( eit = pos.tpPieces.begin(); eit != eend; eit++ )
	{
		if ( !tpBitcodeTable.GetEntry(eit->atomics, eit->type, value) )
		{
			temp = 0;
			d = &currentConfigs[eit->type];
			cit = d->begin(); cend = d->end();
			while ( cit < cend )
			{
				if ( cit->IsActive(eit->atomics) ) 
				{
					temp += cit->weight;
					cit += cit->Jump(eit->atomics);
				}
				else cit += cit->jump;
			}
			value += temp;
			tpBitcodeTable.AddEntry(eit->atomics, eit->type, temp);
		}
	} 

	eend = pos.iePieces.end();
	for( eit = pos.iePieces.begin(); eit != eend; eit++ )
	{
		if ( !ieBitcodeTable.GetEntry(eit->atomics, eit->type, value) )
		{
			temp = 0;
			d = &currentConfigs[eit->type];
			cit = d->begin(); cend = d->end();
			while ( cit < cend )
			{
				if ( cit->IsActive(eit->atomics) ) 
				{
					temp -= cit->weight;
					cit += cit->Jump(eit->atomics);
				}
				else cit += cit->jump;
			}
			value += temp;
			ieBitcodeTable.AddEntry(eit->atomics, eit->type, temp);
		}
	} 

	return value;
}

float Evaluation::GetExampleValue(Position& pos, double& posNGradCond)
{
ConfigurationSet* d;
ConfigurationSetIt cend, cit;
PiecePropertiesSetIt eend, eit;
float value = 0;
	
	posNGradCond = 0;

	// Calculate position value
	eend = pos.tpPieces.end();
	for( eit = pos.tpPieces.begin(); eit != eend; eit++ )
	{
		d = &currentConfigs[eit->type];
		cit = d->begin(); cend = d->end();
		while ( cit < cend )
		{
			if ( cit->IsActive(eit->atomics) ) 
			{
				value += cit->weight;
				posNGradCond += cit->ngrad1 * cit->preconditioner;
				cit += cit->Jump(eit->atomics);
			}
			else cit += cit->jump;
		}
	} 

	eend = pos.iePieces.end();
	for( eit = pos.iePieces.begin(); eit != eend; eit++ )
	{
		d = &currentConfigs[eit->type];
		cit = d->begin(); cend = d->end();
		while ( cit < cend )
		{
			if ( cit->IsActive(eit->atomics) ) 
			{
				value -= cit->weight;
				posNGradCond -= cit->ngrad1 * cit->preconditioner;
				cit += cit->Jump(eit->atomics);
			}
			else cit += cit->jump;
		}
	} 

	return value;
}

float Evaluation::BuildPreconditioner(Position& pos)
{
ConfigurationSet* d;
ConfigurationSetIt cend, cit;
PiecePropertiesSetIt eend, eit;
float value = 0;

	// Count the configurations
	eend = pos.tpPieces.end();
	for( eit = pos.tpPieces.begin(); eit != eend; eit++ )
	{
		d = &currentConfigs[eit->type];
		cit = d->begin(); cend = d->end();
		while ( cit < cend )
		{
			if ( cit->IsActive(eit->atomics) ) 
			{
				value += cit->weight;
				cit->count++;
				cit += cit->Jump(eit->atomics);
			}
			else cit += cit->jump;
		}
	} 

	eend = pos.iePieces.end();
	for( eit = pos.iePieces.begin(); eit != eend; eit++ )
	{
		d = &currentConfigs[eit->type];
		cit = d->begin(); cend = d->end();
		while ( cit < cend )
		{
			if ( cit->IsActive(eit->atomics) ) 
			{
				value -= cit->weight;
				cit->count--;
				cit += cit->Jump(eit->atomics);
			}
			else cit += cit->jump;
		}
	} 

	// Update preconditioner
	eend = pos.tpPieces.end();
	for( eit = pos.tpPieces.begin(); eit != eend; eit++ )
	{
		d = &currentConfigs[eit->type];
		cit = d->begin(); cend = d->end();
		while ( cit < cend )
		{
			if ( cit->IsActive(eit->atomics) ) 
			{
				if ( cit->count != 0 ) cit->preconditioner += cit->count * cit->count;
				cit->count = 0;
				cit += cit->Jump(eit->atomics);
			}
			else cit += cit->jump;
		}
	} 

	eend = pos.iePieces.end();
	for( eit = pos.iePieces.begin(); eit != eend; eit++ )
	{
		d = &currentConfigs[eit->type];
		cit = d->begin(); cend = d->end();
		while ( cit < cend )
		{
			if ( cit->IsActive(eit->atomics) ) 
			{
				if ( cit->count != 0 ) cit->preconditioner += cit->count * cit->count;
				cit->count = 0;
				cit += cit->Jump(eit->atomics);
			}
			else cit += cit->jump;
		}
	} 

	return value;
}

void Evaluation::UpdateGradient(Position& pos, double error)
{
ConfigurationSet* d;
ConfigurationSetIt cend, cit;
PiecePropertiesSetIt estart, eend, eit;

	estart = pos.tpPieces.begin(); eend = pos.tpPieces.end();
	for( eit = estart; eit != eend; eit++ )
	{
		d = &currentConfigs[eit->type];

		cit = d->begin(); cend = d->end();
		while ( cit < cend )
		{
			if ( cit->IsActive(eit->atomics) ) 
			{
				cit->ngrad1 += error;
				cit += cit->Jump(eit->atomics);
			}
			else cit += cit->jump;

		}
	} 

	estart = pos.iePieces.begin(); eend = pos.iePieces.end();
	for( eit = estart; eit != eend; eit++ )
	{
		d = &currentConfigs[eit->type];
	
		cit = d->begin(); cend = d->end();
		while ( cit < cend )
		{
			if ( cit->IsActive(eit->atomics) ) 
			{	
				cit->ngrad1 -= error;
				cit += cit->Jump(eit->atomics);
			}
			else cit += cit->jump;

		}
	} 
}

double Evaluation::GetAverageError(PositionSet& set)
{
double error = 0;
PositionSetIt pstart, pend, pit;

	// Loop over all positions
	pstart = set.begin(); pend = set.end();

	for( pit = pstart; pit != pend; pit++ )
		error += fabs(pit->value - GetExampleValue(*pit));
	
	if ( set.size() == 0 )
		return 0;
	else
		return error / ((double) set.size());
}

float Evaluation::GetHitPercentage(PositionSet& set)
{
int nOfHits = 0;
PositionSetIt pstart, pend, pit;

	// Loop over all positions
	pstart = set.begin(); pend = set.end();

	for( pit = pstart; pit != pend; pit++ )
		if ( fabs(pit->value - GetExampleValue(*pit)) < 1.0 ) nOfHits++;
	
	if ( set.size() == 0 )
		return 100;
	else
		return 100 * nOfHits / ((float) set.size());
}

void Evaluation::RemoveConfigurations(int piece)
{
	currentConfigs[piece].clear();
}

int Evaluation::AddConfiguration(int piece)
{
Configuration c;
	currentConfigs[piece].push_back(c);
	return currentConfigs[piece].size() - 1;
}

Configuration& Evaluation::GetConfiguration(int piece, int nr)
{
	return currentConfigs[piece][nr];
}

void Evaluation::OptimizeConfigurationSet()
{
ConfigurationSet tempConfigs;
ConfigurationSetIt it, end;
int type, piece;

	for ( type = 0; type < MAX_POSITIONTYPES; type++ )
	{
		for ( piece = 0; piece < 6; piece++ )
		{
			configs[type][piece].swap(tempConfigs);
			it = tempConfigs.begin(); 
			end = tempConfigs.end();
			while ( it < end )
			{
				RemoveUnimportantConfigurations(it, configs[type][piece]); 
			}

			cout << "[" << type << "," << piece << "]: " << tempConfigs.size() 
				 << "-->" << configs[type][piece].size() << "  ";


			tempConfigs.clear();
		}
	}
}

int Evaluation::RemoveUnimportantConfigurations(ConfigurationSetIt& it, ConfigurationSet& set)
{
int count = 0, temp, cnr;
unsigned int not_msac = 0;
ConfigurationSetIt begin = it, end = it + it->jump, current;

	it++;

	if ( begin->jump == 1 && begin->weight < 0.05 && begin->weight > -0.05 )
	{ 
		// this is 1. a most specific configuration an 2. the configuration is unimportant
		// let's remove it
		return count;
	}
	else
	{
		set.push_back(*begin);
		cnr = set.size() - 1;
		count++;
	}

	while ( it < end )
	{
		current = it;
		temp = RemoveUnimportantConfigurations(it, set);
		if ( temp > 0 )
		{
			count += temp;
			not_msac |= (begin->bitcode ^ current->bitcode);
		}
	}

	if ( count == 1 && begin->weight < 0.05 && begin->weight > -0.05 )
	{
		// this is 1. a most specific configuration an 2. the configuration is unimportant
		// let's remove it
		count--;
		set.erase(set.end() - 1);
	}
	else
	{
		// update jump and not_msac
		Configuration& c = set[cnr];
		c.jump = count;
		c.not_msac = not_msac;
	}

	return count;
}

#endif // EN_VERSION

