/***************************************************************************
                    pgnfile.cpp  -  handling of pgn games
                             -------------------
    begin                : Sat Jun 3 2000
    copyright            : (C) 1996-2001 by Matthias Luescher
    email                : m.luescher@datacomm.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


// disable warning C4786: symbol greater than 255 character,
#ifndef GNU_VERSION
#pragma warning(disable: 4786) // Visual C++ stuff
#endif

#include "pgnfile.h"

#include <string>
#include <fstream>
#include <iostream>

using namespace std;

string uintToString(unsigned int i)
{
	string str;
	char c;
	if ( i == 0 ) return "0";

	while ( i > 0 )
	{

		c = (i % 10) + '0';
		str.insert(str.begin(), c);
		i /= 10;
	}

	return str;
}


string FENTag::SetTag(int* board, bool toPlay, 
	  		    	  bool* castlingAvailability,
					  int halfmoveClock, 
					  int fullmoveNumber, 
					  int epSquare )
{
	char pl[12] = {'p', 'r', 'b', 'n', 'q', 'k',
		 		   'P', 'R', 'B', 'N', 'Q', 'K'};
	int piece;
	unsigned int count;
	tag = "";

	for ( int rank = 7; rank >= 0; rank-- )
	{
		count = 0;
		for ( int column = 0; column < 8; column++ )
		{
			piece = board[column + 8 * rank];
			if ( piece == PGN_EMPTY_SQUARE )
			{
				count++;
			}
			else
			{
				if ( count > 0 ) tag += (char) ('0' + count);
				tag += pl[piece];
				count = 0;
			}
		}

		if ( count > 0 ) tag += (char) ('0' + count);
		if ( rank > 0 ) tag += '/';
	}

	if ( toPlay == PGN_WHITE ) tag += " w ";
	else tag += " b ";

	count = tag.size();
	if ( castlingAvailability != NULL )
	{
		if ( castlingAvailability[PGN_W_KS_CASTLING] ) tag += 'K';
		if ( castlingAvailability[PGN_W_QS_CASTLING] ) tag += 'Q';
		if ( castlingAvailability[PGN_B_KS_CASTLING] ) tag += 'k';
		if ( castlingAvailability[PGN_B_QS_CASTLING] ) tag += 'q';
	}

	if ( count == tag.size() ) tag += "- ";
	else tag += " ";

	if ( epSquare == PGN_NO_SQUARE )
		tag += "- ";
	else
	{
		tag += (char) ((epSquare & 7) + 'a');
		tag += (char) ((epSquare >> 3) + '1');
		tag += " ";
	}

	tag += (uintToString(halfmoveClock) + " ");
	tag += uintToString(fullmoveNumber);

	return tag;
}

// ----------------------------------------------------------
// LoadFile loads a pgn file specified by the string name
// which can contain a directory but must contain a name
// of an existing pgn file.
// One or more files can be loaded.
// ----------------------------------------------------------

bool PGNFile::LoadFile(string name)
{
	fstream file;
	string str, temp;
	char c;
	bool setup = false;
	file.open( name.c_str(), ios::in); 

	if ( !file.good() ) return false;

	file >> c;

	while ( file.good() )
	{
		gameList.push_back(PGNGame());
		PGNGame& g = gameList.back();

		while ( c == ' ' && file.good() ) file >> c;

		while ( c == '[' )
		{
			getline(file, str, ']');
			g.tags.push_back(str);
#ifdef GNU_VERSION
			if ( str.compare("FEN ", 0, 4) == 0 || str.compare("fen ", 0, 4) == 0
				 || str.compare("Fen ", 0, 4) == 0 )
#else
			if ( str.compare(0,4, "FEN ") == 0 || str.compare(0,4, "fen ") == 0
				 || str.compare(0,4, "Fen ") == 0 )
#endif
			{
				setup = g.ParseFenTag(str);
			}
			file >> c;
		}

		if ( !setup ) 
			g.ParseFenTag("FEN \"rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1\"");

		while ( c == ' ' && file.good() ) file >> c;

		while ( c != '[' && file.good() )
		{
			if ( c >= '0' && c <= '9' ) // must be a move number or a result
			{
				str = c;
				file >> c;
				if ( file.good() )
				{
					if ( c >= '0' && c <= '9' ) // must be a move number
					{
						getline(file, temp, '.');
						str += c + temp;
					}
					else if ( c == '.' )
					{
						str += c;
					}
					else // must be a result
					{
						file >> temp;
						str += c + temp;
					}
				}
			}
			else
			{
				switch ( c )
				{
				case '(': 
					getline(file, str, ')');
					str = c + str + ')';
					break;
				case '{': 
					getline(file, str, '}');
					str = c + str + '}';
					break;
				case '<': 
					getline(file, str, '>');
					str = c + str + '>';
					break;
				case '*':
					str = c;
					break;
				default:
					file >> str;
					str = c + str;
					break;
				}
			}

			g.moveText.push_back(str);
			file >> c;
			while ( c == ' ' && file.good() ) file >> c;
		}
	}
	
	file.close();
	return true;
}

// ----------------------------------------------------------
// A PGN-File can contain more than one game.
// GetGame can be used to access a game specified by nr
// The first game is game nr 0
// ----------------------------------------------------------

PGNGame* PGNFile::GetGame(int nr)
{
	if ( nr >= 0 && nr < (int) gameList.size() )
	{
		gameList[nr].Reset();
		return &gameList[nr];
	}
	else
		return NULL;
}

// ----------------------------------------------------------
// GetNextMove 
// The piece (piece) moves from (from_r,from_c) to (to_r,to_c) 
// and eventually promotes to (p_piece).
// r and c stand for rank and columns. Specified rank and 
// columns have values between 0 and 7.
// One or more parameters can be unspecified 
// (#define PGN_NOT_SPECIFIED 8) but they can be
// clearly identified from the given parameters.
// Tip: Use your engine to generate all valid moves in the 
// current position and find the only move that matches all 
// known parameters.
// false is returned if there is no next move.
// ----------------------------------------------------------

bool PGNGame::GetNextMove( int& from_r, int& from_c, int& to_r, 
						   int& to_c, int& piece, int& p_piece )
{
	bool isMove = false;
	string str;

	do // Ignore all move text elements but moves
	{
		if ( currentElement >= (int) moveText.size() ) return false;

		str = moveText[currentElement];
		currentElement++;

		if ( (str[0] >= 'a' && str[0] <= 'h') // pawn moves
			 || str[0] == 'O' // castling move
			 || str[0] == 'P' || str[0] == 'N' || str[0] == 'B' // piece indication
			 || str[0] == 'R' || str[0] == 'Q' || str[0] == 'K' )
			isMove = true;		
	}
	while ( !isMove );

	int pos = str.size() - 1;
	from_r = PGN_NOT_SPECIFIED; from_c = PGN_NOT_SPECIFIED; 
	to_r = PGN_NOT_SPECIFIED; to_c = PGN_NOT_SPECIFIED;
	piece = PGN_PAWN; p_piece = PGN_NO_PIECE;

	if ( str == "O-O" ) // kingside castling
	{
		to_c = 6;
		piece = PGN_KING;
		return true;
	}

	if ( str == "O-O-O" ) // queenside castling
	{
		to_c = 2;
		piece = PGN_KING;
		return true;
	}

	while ( str[pos] == '#' || str[pos] == '+' || str[pos] == '?' || str[pos] == '!' ) pos--;

	switch ( str[pos] )
	{
	case 'Q': p_piece = PGN_QUEEN; pos -= 2; break;
	case 'R': p_piece = PGN_ROOK; pos -= 2; break;
	case 'B': p_piece = PGN_BISHOP; pos -= 2; break;
	case 'N': p_piece = PGN_KNIGHT; pos -= 2; break;
	}

	to_c = (str[pos - 1] - 'a');
	to_r = (str[pos] - '1');
	if ( to_c < 0 || to_c > 7 || to_r < 0 || to_r > 7 ) return false;
	pos -= 2;

	if ( pos < 0 ) return true;
	if ( str[pos] == 'x' ) pos--;

	if ( str[pos] >= '1' && str[pos] <= '8') 
	{
		from_r = str[pos] - '1';
		pos--;
	}
	if ( pos < 0 ) return true;

	if ( str[pos] >= 'a' && str[pos] <= 'h') 
	{
		from_c = str[pos] - 'a';
		pos--;
	}
	if ( pos < 0 ) return true;

	switch ( str[pos] )
	{
	case 'Q': piece = PGN_QUEEN; break;
	case 'R': piece = PGN_ROOK; break;
	case 'B': piece = PGN_BISHOP; break;
	case 'N': piece = PGN_KNIGHT; break;
	case 'K': piece = PGN_KING; break;
	}

	return true;
}

bool PGNGame::UndoLastMove()
{
	// some more work is needed
	return true;
}

bool PGNGame::ParseFenTag(string str)
{
	int pos = 0, rank = 7, column = 0, size = str.size();
	
	while ( str[pos] != '"' && pos < size ) pos++;
	pos++;

	do
	{
		switch ( str[pos] )
		{
		case 'P': pieces[PGN_W_PAWN].push_back(rank * 8 + column); column++; break;
		case 'N': pieces[PGN_W_KNIGHT].push_back(rank * 8 + column); column++; break;
		case 'B': pieces[PGN_W_BISHOP].push_back(rank * 8 + column); column++; break;
		case 'R': pieces[PGN_W_ROOK].push_back(rank * 8 + column); column++; break;
		case 'Q': pieces[PGN_W_QUEEN].push_back(rank * 8 + column); column++; break;
		case 'K': pieces[PGN_W_KING].push_back(rank * 8 + column); column++; break;
		case 'p': pieces[PGN_B_PAWN].push_back(rank * 8 + column); column++; break;
		case 'n': pieces[PGN_B_KNIGHT].push_back(rank * 8 + column); column++; break;
		case 'b': pieces[PGN_B_BISHOP].push_back(rank * 8 + column); column++; break;
		case 'r': pieces[PGN_B_ROOK].push_back(rank * 8 + column); column++; break;
		case 'q': pieces[PGN_B_QUEEN].push_back(rank * 8 + column); column++; break;
		case 'k': pieces[PGN_B_KING].push_back(rank * 8 + column); column++; break;
		}

		if ( str[pos] >= '1' && str[pos] <= '8' ) column += (str[pos] - '0');
		
		if ( column >= 8 || str[pos] == '/' )
		{
			if ( str[pos] != '/' ) pos++;
			rank--;
			column = 0;
		}
		pos++;
	}
	while ( rank >= 0  && pos < size );

	if ( pieces[PGN_W_KING].size() != 1 || pieces[PGN_B_KING].size() != 1 ) 
	{
		for ( int i = 0; i < 12; i++ ) pieces[i].clear(); 
		return false;
	}

	while ( str[pos] == ' ' && pos < size ) pos++;

	if ( str[pos] == 'b' ) activeColor = PGN_BLACK;
	else activeColor = PGN_WHITE;
	pos++;

	while ( str[pos] == ' ' && pos < size ) pos++;

	for ( int i = 0; i < 4; i++ )
		castlingAvailability[i] = false;

	while ( str[pos] != ' ' && pos < size )
	{
		switch ( str[pos] )
		{
		case 'K': castlingAvailability[PGN_W_KS_CASTLING] = true; break;
		case 'Q': castlingAvailability[PGN_W_QS_CASTLING] = true; break;
		case 'k': castlingAvailability[PGN_B_KS_CASTLING] = true; break;
		case 'q': castlingAvailability[PGN_B_QS_CASTLING] = true; break;
		}
		pos++;
	}

	while ( str[pos] == ' ' && pos < size ) pos++;

	if ( str[pos] == '-' ) 
	{
		epSquare = PGN_NO_SQUARE;
		pos++;
	}
	else
	{
		epSquare = (str[pos] - 'a') + 8 * (str[pos + 1] - '1');
		pos += 2;
	}

	while ( str[pos] == ' ' && pos < size ) pos++;

	halfmoveClock = 0;
	while ( str[pos] != ' ' && pos < size ) 
	{
		halfmoveClock = 10 * halfmoveClock + (str[pos] - '0');
		pos++;
	}

	while ( str[pos] == ' ' && pos < size ) pos++;

	fullmoveNumber = 0;
	while ( str[pos] != ' ' && str[pos] != '"' && pos < size ) 
	{
		fullmoveNumber = 10 * fullmoveNumber + (str[pos] - '0');
		pos++;
	}
	return true;
}

int PGNGame::GetResult()
{
	int pos = moveText.size() - 1;
	while ( pos >= 0 )
	{
		if ( moveText[pos] == "1-0" ) return PGN_WHITE_WINS;
		if ( moveText[pos] == "0-1" ) return PGN_BLACK_WINS;
		if ( moveText[pos] == "1/2-1/2" ) return PGN_DRAW;
		if ( moveText[pos] == "*" ) return PGN_RESULT_UNKNOWN;
		pos--;
	}
	return PGN_RESULT_UNKNOWN;
}



